@@ -5,10 +5,12 @@
 #include <wlr/config.h>
 #include <wlr/xwayland.h>
 #include <xcb/render.h>
-#if HAS_XCB_ERRORS
+#include "config.h"
+#include "xwayland/selection.h"
+
+#if HAVE_XCB_ERRORS
 #include <xcb/xcb_errors.h>
 #endif
-#include "xwayland/selection.h"
 
 /* This is in xcb/xcb_event.h, but pulling xcb-util just for a constant
  * others redefine anyway is meh
@@ -17,6 +19,7 @@
 
 enum atom_name {
 	WL_SURFACE_ID,
+	WL_SURFACE_SERIAL,
 	WM_DELETE_WINDOW,
 	WM_PROTOCOLS,
 	WM_HINTS,
@@ -31,6 +34,7 @@ enum atom_name {
 	NET_WM_PID,
 	NET_WM_NAME,
 	NET_WM_STATE,
+	NET_WM_STRUT_PARTIAL,
 	NET_WM_WINDOW_TYPE,
 	WM_TAKE_FOCUS,
 	WINDOW,
@@ -84,11 +88,10 @@ enum atom_name {
 	DND_ACTION_PRIVATE,
 	NET_CLIENT_LIST,
 	NET_CLIENT_LIST_STACKING,
+	NET_WORKAREA,
 	ATOM_LAST // keep last
 };
 
-extern const char *const atom_map[ATOM_LAST];
-
 struct wlr_xwm {
 	struct wlr_xwayland *xwayland;
 	struct wl_event_source *event_source;
@@ -111,10 +114,10 @@ struct wlr_xwm {
 	struct wlr_xwayland_surface *focus_surface;
 
 	// Surfaces in creation order
-	struct wl_list surfaces; // wlr_xwayland_surface::link
+	struct wl_list surfaces; // wlr_xwayland_surface.link
 	// Surfaces in bottom-to-top stacking order, for _NET_CLIENT_LIST_STACKING
-	struct wl_list surfaces_in_stack_order; // wlr_xwayland_surface::stack_link
-	struct wl_list unpaired_surfaces; // wlr_xwayland_surface::unpaired_link
+	struct wl_list surfaces_in_stack_order; // wlr_xwayland_surface.stack_link
+	struct wl_list unpaired_surfaces; // wlr_xwayland_surface.unpaired_link
 	struct wl_list pending_startup_ids; // pending_startup_id
 
 	struct wlr_drag *drag;
@@ -123,13 +126,14 @@ struct wlr_xwm {
 	const xcb_query_extension_reply_t *xfixes;
 	const xcb_query_extension_reply_t *xres;
 	uint32_t xfixes_major_version;
-#if HAS_XCB_ERRORS
+#if HAVE_XCB_ERRORS
 	xcb_errors_context_t *errors_context;
 #endif
 	unsigned int last_focus_seq;
 
 	struct wl_listener compositor_new_surface;
 	struct wl_listener compositor_destroy;
+	struct wl_listener shell_v1_new_surface;
 	struct wl_listener seat_set_selection;
 	struct wl_listener seat_set_primary_selection;
 	struct wl_listener seat_start_drag;
@@ -157,4 +161,8 @@ char *xwm_get_atom_name(struct wlr_xwm *xwm, xcb_atom_t atom);
 bool xwm_atoms_contains(struct wlr_xwm *xwm, xcb_atom_t *atoms,
 	size_t num_atoms, enum atom_name needle);
 
+xcb_void_cookie_t xwm_send_event_with_size(xcb_connection_t *c,
+	uint8_t propagate, xcb_window_t destination,
+	uint32_t event_mask, const void *event, uint32_t length);
+
 #endif
