@@ -22,6 +22,40 @@ static void resource_handle_destroy(struct wl_client *client,
 static const struct zwlr_layer_shell_v1_interface layer_shell_implementation;
 static const struct zwlr_layer_surface_v1_interface layer_surface_implementation;
 
+static void layer_surface_configure_destroy(
+		struct wlr_layer_surface_v1_configure *configure) {
+	if (configure == NULL) {
+		return;
+	}
+	wl_list_remove(&configure->link);
+	free(configure);
+}
+
+static void layer_surface_reset(struct wlr_layer_surface_v1 *surface) {
+	surface->configured = false;
+	surface->initialized = false;
+
+	struct wlr_xdg_popup *popup, *popup_tmp;
+	wl_list_for_each_safe(popup, popup_tmp, &surface->popups, link) {
+		wlr_xdg_popup_destroy(popup);
+	}
+
+	struct wlr_layer_surface_v1_configure *configure, *tmp;
+	wl_list_for_each_safe(configure, tmp, &surface->configure_list, link) {
+		layer_surface_configure_destroy(configure);
+	}
+}
+
+static void layer_surface_destroy(struct wlr_layer_surface_v1 *surface) {
+	wlr_surface_unmap(surface->surface);
+	layer_surface_reset(surface);
+
+	wl_signal_emit_mutable(&surface->events.destroy, surface);
+	wl_resource_set_user_data(surface->resource, NULL);
+	free(surface->namespace);
+	free(surface);
+}
+
 static struct wlr_layer_shell_v1 *layer_shell_from_resource(
 		struct wl_resource *resource) {
 	assert(wl_resource_instance_of(resource, &zwlr_layer_shell_v1_interface,
@@ -38,23 +72,12 @@ struct wlr_layer_surface_v1 *wlr_layer_surface_v1_from_resource(
 
 static const struct wlr_surface_role layer_surface_role;
 
-bool wlr_surface_is_layer_surface(struct wlr_surface *surface) {
-	return surface->role == &layer_surface_role;
-}
-
-struct wlr_layer_surface_v1 *wlr_layer_surface_v1_from_wlr_surface(
+struct wlr_layer_surface_v1 *wlr_layer_surface_v1_try_from_wlr_surface(
 		struct wlr_surface *surface) {
-	assert(wlr_surface_is_layer_surface(surface));
-	return (struct wlr_layer_surface_v1 *)surface->role_data;
-}
-
-static void layer_surface_configure_destroy(
-		struct wlr_layer_surface_v1_configure *configure) {
-	if (configure == NULL) {
-		return;
+	if (surface->role != &layer_surface_role || surface->role_resource == NULL) {
+		return NULL;
 	}
-	wl_list_remove(&configure->link);
-	free(configure);
+	return wlr_layer_surface_v1_from_resource(surface->role_resource);
 }
 
 static void layer_surface_handle_ack_configure(struct wl_client *client,
@@ -107,6 +130,12 @@ static void layer_surface_handle_set_size(struct wl_client *client,
 		return;
 	}
 
+	if (width > INT32_MAX || height > INT32_MAX) {
+		wl_client_post_implementation_error(client,
+			"zwlr_layer_surface_v1.set_size: width and height can't be greater than INT32_MAX");
+		return;
+	}
+
 	if (surface->current.desired_width == width
 			&& surface->current.desired_height == height) {
 		surface->pending.committed &= ~WLR_LAYER_SURFACE_V1_STATE_DESIRED_SIZE;
@@ -222,6 +251,10 @@ static void layer_surface_handle_get_popup(struct wl_client *client,
 	if (!parent) {
 		return;
 	}
+	if (popup->parent != NULL) {
+		wl_resource_post_error(layer_resource, -1, "xdg_popup already has a parent");
+		return;
+	}
 	popup->parent = parent->surface;
 	wl_list_insert(&parent->popups, &popup->link);
 	wl_signal_emit_mutable(&parent->events.new_popup, popup);
@@ -262,37 +295,16 @@ static const struct zwlr_layer_surface_v1_interface layer_surface_implementation
 	.set_layer = layer_surface_set_layer,
 };
 
-static void layer_surface_unmap(struct wlr_layer_surface_v1 *surface) {
-	surface->configured = surface->mapped = false;
-
-	// TODO: probably need to ungrab before this event
-	wl_signal_emit_mutable(&surface->events.unmap, surface);
-
-	struct wlr_xdg_popup *popup, *popup_tmp;
-	wl_list_for_each_safe(popup, popup_tmp, &surface->popups, link) {
-		wlr_xdg_popup_destroy(popup);
-	}
-
-	struct wlr_layer_surface_v1_configure *configure, *tmp;
-	wl_list_for_each_safe(configure, tmp, &surface->configure_list, link) {
-		layer_surface_configure_destroy(configure);
-	}
-}
-
-static void layer_surface_resource_destroy(struct wl_resource *resource) {
-	struct wlr_layer_surface_v1 *surface =
-		wlr_layer_surface_v1_from_resource(resource);
-	if (surface != NULL) {
-		wlr_surface_destroy_role_object(surface->surface);
-	}
-}
-
 uint32_t wlr_layer_surface_v1_configure(struct wlr_layer_surface_v1 *surface,
 		uint32_t width, uint32_t height) {
+	if (!surface->initialized) {
+		wlr_log(WLR_ERROR, "A configure is sent to an uninitialized wlr_layer_surface_v1 %p",
+			surface);
+	}
+
 	struct wl_display *display =
 		wl_client_get_display(wl_resource_get_client(surface->resource));
-	struct wlr_layer_surface_v1_configure *configure =
-		calloc(1, sizeof(struct wlr_layer_surface_v1_configure));
+	struct wlr_layer_surface_v1_configure *configure = calloc(1, sizeof(*configure));
 	if (configure == NULL) {
 		wl_client_post_no_memory(wl_resource_get_client(surface->resource));
 		return surface->pending.configure_serial;
@@ -308,13 +320,26 @@ uint32_t wlr_layer_surface_v1_configure(struct wlr_layer_surface_v1 *surface,
 }
 
 void wlr_layer_surface_v1_destroy(struct wlr_layer_surface_v1 *surface) {
+	if (surface == NULL) {
+		return;
+	}
 	zwlr_layer_surface_v1_send_closed(surface->resource);
-	wlr_surface_destroy_role_object(surface->surface);
+	layer_surface_destroy(surface);
 }
 
 static void layer_surface_role_commit(struct wlr_surface *wlr_surface) {
 	struct wlr_layer_surface_v1 *surface =
-		wlr_layer_surface_v1_from_wlr_surface(wlr_surface);
+		wlr_layer_surface_v1_try_from_wlr_surface(wlr_surface);
+	if (surface == NULL) {
+		return;
+	}
+
+	if (wlr_surface_has_buffer(surface->surface) && !surface->configured) {
+		wl_resource_post_error(surface->resource,
+			ZWLR_LAYER_SHELL_V1_ERROR_ALREADY_CONSTRUCTED,
+			"layer_surface has never been configured");
+		return;
+	}
 
 	const uint32_t horiz = ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT |
 		ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
@@ -336,63 +361,45 @@ static void layer_surface_role_commit(struct wlr_surface *wlr_surface) {
 		return;
 	}
 
+	if (surface->surface->unmap_commit) {
+		layer_surface_reset(surface);
+
+		assert(!surface->initialized);
+		surface->initial_commit = false;
+	} else {
+		surface->initial_commit = !surface->initialized;
+		surface->initialized = true;
+	}
+
 	surface->current = surface->pending;
 	surface->pending.committed = 0;
 
-	if (wlr_surface_has_buffer(surface->surface) && !surface->configured) {
-		wl_resource_post_error(surface->resource,
-			ZWLR_LAYER_SHELL_V1_ERROR_ALREADY_CONSTRUCTED,
-			"layer_surface has never been configured");
-		return;
-	}
-
 	if (!surface->added) {
 		surface->added = true;
-		assert(!surface->configured);
-		assert(!surface->mapped);
 		wl_signal_emit_mutable(&surface->shell->events.new_surface, surface);
 		// Return early here as the compositor may have closed this layer surface
 		// in response to the new_surface event.
 		return;
 	}
 
-	if (surface->configured && wlr_surface_has_buffer(surface->surface) &&
-			!surface->mapped) {
-		surface->mapped = true;
-		wl_signal_emit_mutable(&surface->events.map, surface);
-	}
-}
-
-static void layer_surface_role_precommit(struct wlr_surface *wlr_surface,
-		const struct wlr_surface_state *state) {
-	struct wlr_layer_surface_v1 *surface =
-		wlr_layer_surface_v1_from_wlr_surface(wlr_surface);
-
-	if (state->committed & WLR_SURFACE_STATE_BUFFER && state->buffer == NULL) {
-		// This is a NULL commit
-		if (surface->configured && surface->mapped) {
-			layer_surface_unmap(surface);
-		}
+	if (wlr_surface_has_buffer(wlr_surface)) {
+		wlr_surface_map(wlr_surface);
 	}
 }
 
 static void layer_surface_role_destroy(struct wlr_surface *wlr_surface) {
 	struct wlr_layer_surface_v1 *surface =
-		wlr_layer_surface_v1_from_wlr_surface(wlr_surface);
-
-	if (surface->configured && surface->mapped) {
-		layer_surface_unmap(surface);
+		wlr_layer_surface_v1_try_from_wlr_surface(wlr_surface);
+	if (surface == NULL) {
+		return;
 	}
-	wl_signal_emit_mutable(&surface->events.destroy, surface);
-	wl_resource_set_user_data(surface->resource, NULL);
-	free(surface->namespace);
-	free(surface);
+
+	layer_surface_destroy(surface);
 }
 
 static const struct wlr_surface_role layer_surface_role = {
 	.name = "zwlr_layer_surface_v1",
 	.commit = layer_surface_role_commit,
-	.precommit = layer_surface_role_precommit,
 	.destroy = layer_surface_role_destroy,
 };
 
@@ -406,14 +413,13 @@ static void layer_shell_handle_get_layer_surface(struct wl_client *wl_client,
 	struct wlr_surface *wlr_surface =
 		wlr_surface_from_resource(surface_resource);
 
-	struct wlr_layer_surface_v1 *surface =
-		calloc(1, sizeof(struct wlr_layer_surface_v1));
+	struct wlr_layer_surface_v1 *surface = calloc(1, sizeof(*surface));
 	if (surface == NULL) {
 		wl_client_post_no_memory(wl_client);
 		return;
 	}
 
-	if (!wlr_surface_set_role(wlr_surface, &layer_surface_role, surface,
+	if (!wlr_surface_set_role(wlr_surface, &layer_surface_role,
 			client_resource, ZWLR_LAYER_SHELL_V1_ERROR_ROLE)) {
 		free(surface);
 		return;
@@ -453,14 +459,14 @@ static void layer_shell_handle_get_layer_surface(struct wl_client *wl_client,
 	wl_list_init(&surface->popups);
 
 	wl_signal_init(&surface->events.destroy);
-	wl_signal_init(&surface->events.map);
-	wl_signal_init(&surface->events.unmap);
 	wl_signal_init(&surface->events.new_popup);
 
 	wlr_log(WLR_DEBUG, "new layer_surface %p (res %p)",
 			surface, surface->resource);
 	wl_resource_set_implementation(surface->resource,
-		&layer_surface_implementation, surface, layer_surface_resource_destroy);
+		&layer_surface_implementation, surface, NULL);
+
+	wlr_surface_set_role_object(wlr_surface, surface->resource);
 }
 
 static const struct zwlr_layer_shell_v1_interface layer_shell_implementation = {
@@ -471,7 +477,6 @@ static const struct zwlr_layer_shell_v1_interface layer_shell_implementation = {
 static void layer_shell_bind(struct wl_client *wl_client, void *data,
 		uint32_t version, uint32_t id) {
 	struct wlr_layer_shell_v1 *layer_shell = data;
-	assert(wl_client && layer_shell);
 
 	struct wl_resource *resource = wl_resource_create(
 			wl_client, &zwlr_layer_shell_v1_interface, version, id);
@@ -492,16 +497,17 @@ static void handle_display_destroy(struct wl_listener *listener, void *data) {
 	free(layer_shell);
 }
 
-struct wlr_layer_shell_v1 *wlr_layer_shell_v1_create(struct wl_display *display) {
-	struct wlr_layer_shell_v1 *layer_shell =
-		calloc(1, sizeof(struct wlr_layer_shell_v1));
+struct wlr_layer_shell_v1 *wlr_layer_shell_v1_create(struct wl_display *display,
+		uint32_t version) {
+	assert(version <= LAYER_SHELL_VERSION);
+
+	struct wlr_layer_shell_v1 *layer_shell = calloc(1, sizeof(*layer_shell));
 	if (!layer_shell) {
 		return NULL;
 	}
 
 	struct wl_global *global = wl_global_create(display,
-		&zwlr_layer_shell_v1_interface, LAYER_SHELL_VERSION,
-		layer_shell, layer_shell_bind);
+		&zwlr_layer_shell_v1_interface, version, layer_shell, layer_shell_bind);
 	if (!global) {
 		free(layer_shell);
 		return NULL;
@@ -537,10 +543,10 @@ void wlr_layer_surface_v1_for_each_surface(struct wlr_layer_surface_v1 *surface,
 }
 
 void wlr_layer_surface_v1_for_each_popup_surface(struct wlr_layer_surface_v1 *surface,
-		wlr_surface_iterator_func_t iterator, void *user_data){
+		wlr_surface_iterator_func_t iterator, void *user_data) {
 	struct wlr_xdg_popup *popup;
 	wl_list_for_each(popup, &surface->popups, link) {
-		if (!popup->base->configured || !popup->base->mapped) {
+		if (!popup->base->surface->mapped) {
 			continue;
 		}
 
@@ -575,7 +581,7 @@ struct wlr_surface *wlr_layer_surface_v1_popup_surface_at(
 		double *sub_x, double *sub_y) {
 	struct wlr_xdg_popup *popup;
 	wl_list_for_each(popup, &surface->popups, link) {
-		if (!popup->base->mapped) {
+		if (!popup->base->surface->mapped) {
 			continue;
 		}
 
