@@ -1,23 +1,14 @@
 #include <assert.h>
 #include <stdlib.h>
+#include <wlr/interfaces/wlr_pointer.h>
 #include <wlr/types/wlr_seat.h>
 #include <wlr/types/wlr_virtual_pointer_v1.h>
 #include <wlr/types/wlr_pointer.h>
 #include <wlr/util/log.h>
-#include "util/signal.h"
 #include "wlr-virtual-pointer-unstable-v1-protocol.h"
 
-static void input_device_destroy(struct wlr_input_device *dev) {
-	struct wlr_virtual_pointer_v1 *pointer =
-		(struct wlr_virtual_pointer_v1 *)dev;
-	wl_resource_set_user_data(pointer->resource, NULL);
-	wlr_signal_emit_safe(&pointer->events.destroy, pointer);
-	wl_list_remove(&pointer->link);
-	free(pointer);
-}
-
-static const struct wlr_input_device_impl input_device_impl = {
-	.destroy = input_device_destroy
+static const struct wlr_pointer_impl pointer_impl = {
+	.name = "virtual-pointer",
 };
 
 static const struct zwlr_virtual_pointer_v1_interface virtual_pointer_impl;
@@ -37,16 +28,15 @@ static void virtual_pointer_motion(struct wl_client *client,
 	if (pointer == NULL) {
 		return;
 	}
-	struct wlr_input_device *wlr_dev = &pointer->input_device;
-	struct wlr_event_pointer_motion event = {
-		.device = wlr_dev,
+	struct wlr_pointer_motion_event event = {
+		.pointer = &pointer->pointer,
 		.time_msec = time,
 		.delta_x = wl_fixed_to_double(dx),
 		.delta_y = wl_fixed_to_double(dy),
 		.unaccel_dx = wl_fixed_to_double(dx),
 		.unaccel_dy = wl_fixed_to_double(dy),
 	};
-	wlr_signal_emit_safe(&wlr_dev->pointer->events.motion, &event);
+	wl_signal_emit_mutable(&pointer->pointer.events.motion, &event);
 }
 
 static void virtual_pointer_motion_absolute(struct wl_client *client,
@@ -60,14 +50,13 @@ static void virtual_pointer_motion_absolute(struct wl_client *client,
 	if (x_extent == 0 || y_extent == 0) {
 		return;
 	}
-	struct wlr_input_device *wlr_dev = &pointer->input_device;
-	struct wlr_event_pointer_motion_absolute event = {
-		.device = wlr_dev,
+	struct wlr_pointer_motion_absolute_event event = {
+		.pointer = &pointer->pointer,
 		.time_msec = time,
 		.x = (double)x / x_extent,
 		.y = (double)y / y_extent,
 	};
-	wlr_signal_emit_safe(&wlr_dev->pointer->events.motion_absolute, &event);
+	wl_signal_emit_mutable(&pointer->pointer.events.motion_absolute, &event);
 }
 
 static void virtual_pointer_button(struct wl_client *client,
@@ -78,14 +67,13 @@ static void virtual_pointer_button(struct wl_client *client,
 	if (pointer == NULL) {
 		return;
 	}
-	struct wlr_input_device *wlr_dev = &pointer->input_device;
-	struct wlr_event_pointer_button event = {
-		.device = wlr_dev,
+	struct wlr_pointer_button_event event = {
+		.pointer = &pointer->pointer,
 		.time_msec = time,
 		.button = button,
 		.state = state ? WLR_BUTTON_PRESSED : WLR_BUTTON_RELEASED
 	};
-	wlr_signal_emit_safe(&wlr_dev->pointer->events.button, &event);
+	wl_signal_emit_mutable(&pointer->pointer.events.button, &event);
 }
 
 static void virtual_pointer_axis(struct wl_client *client,
@@ -94,7 +82,7 @@ static void virtual_pointer_axis(struct wl_client *client,
 	if (axis > WL_POINTER_AXIS_HORIZONTAL_SCROLL) {
 		wl_resource_post_error(resource,
 				ZWLR_VIRTUAL_POINTER_V1_ERROR_INVALID_AXIS,
-				"Invalid enumeration value %d", axis);
+				"Invalid enumeration value %" PRIu32, axis);
 		return;
 	}
 	struct wlr_virtual_pointer_v1 *pointer =
@@ -102,9 +90,9 @@ static void virtual_pointer_axis(struct wl_client *client,
 	if (pointer == NULL) {
 		return;
 	}
-	struct wlr_input_device *wlr_dev = &pointer->input_device;
-	pointer->axis_valid[axis] = true;
-	pointer->axis_event[pointer->axis].device = wlr_dev;
+	pointer->axis = axis;
+	pointer->axis_valid[pointer->axis] = true;
+	pointer->axis_event[pointer->axis].pointer = &pointer->pointer;
 	pointer->axis_event[pointer->axis].time_msec = time;
 	pointer->axis_event[pointer->axis].orientation = axis;
 	pointer->axis_event[pointer->axis].delta = wl_fixed_to_double(value);
@@ -117,21 +105,20 @@ static void virtual_pointer_frame(struct wl_client *client,
 	if (pointer == NULL) {
 		return;
 	}
-	struct wlr_input_device *wlr_dev = &pointer->input_device;
 
 	for (size_t i = 0;
 			i < sizeof(pointer->axis_valid) / sizeof(pointer->axis_valid[0]);
 			++i) {
 		if (pointer->axis_valid[i]) {
 			/* Deliver pending axis event */
-			wlr_signal_emit_safe(&wlr_dev->pointer->events.axis,
+			wl_signal_emit_mutable(&pointer->pointer.events.axis,
 					&pointer->axis_event[i]);
-			memset(&pointer->axis_event[i], 0, sizeof(pointer->axis_event[i]));
+			pointer->axis_event[i] = (struct wlr_pointer_axis_event){0};
 			pointer->axis_valid[i] = false;
 		}
 	}
 
-	wlr_signal_emit_safe(&wlr_dev->pointer->events.frame, wlr_dev->pointer);
+	wl_signal_emit_mutable(&pointer->pointer.events.frame, &pointer->pointer);
 }
 
 static void virtual_pointer_axis_source(struct wl_client *client,
@@ -139,7 +126,7 @@ static void virtual_pointer_axis_source(struct wl_client *client,
 	if (source > WL_POINTER_AXIS_SOURCE_WHEEL_TILT) {
 		wl_resource_post_error(resource,
 				ZWLR_VIRTUAL_POINTER_V1_ERROR_INVALID_AXIS_SOURCE,
-				"Invalid enumeration value %d", source);
+				"Invalid enumeration value %" PRIu32, source);
 		return;
 	}
 	struct wlr_virtual_pointer_v1 *pointer =
@@ -147,8 +134,7 @@ static void virtual_pointer_axis_source(struct wl_client *client,
 	if (pointer == NULL) {
 		return;
 	}
-	struct wlr_input_device *wlr_dev = &pointer->input_device;
-	pointer->axis_event[pointer->axis].device = wlr_dev;
+	pointer->axis_event[pointer->axis].pointer = &pointer->pointer;
 	pointer->axis_event[pointer->axis].source = source;
 }
 
@@ -157,7 +143,7 @@ static void virtual_pointer_axis_stop(struct wl_client *client,
 	if (axis > WL_POINTER_AXIS_HORIZONTAL_SCROLL) {
 		wl_resource_post_error(resource,
 				ZWLR_VIRTUAL_POINTER_V1_ERROR_INVALID_AXIS,
-				"Invalid enumeration value %d", axis);
+				"Invalid enumeration value %" PRIu32, axis);
 		return;
 	}
 	struct wlr_virtual_pointer_v1 *pointer =
@@ -165,10 +151,9 @@ static void virtual_pointer_axis_stop(struct wl_client *client,
 	if (pointer == NULL) {
 		return;
 	}
-	struct wlr_input_device *wlr_dev = &pointer->input_device;
 	pointer->axis = axis;
 	pointer->axis_valid[pointer->axis] = true;
-	pointer->axis_event[pointer->axis].device = wlr_dev;
+	pointer->axis_event[pointer->axis].pointer = &pointer->pointer;
 	pointer->axis_event[pointer->axis].time_msec = time;
 	pointer->axis_event[pointer->axis].orientation = axis;
 	pointer->axis_event[pointer->axis].delta = 0;
@@ -181,7 +166,7 @@ static void virtual_pointer_axis_discrete(struct wl_client *client,
 	if (axis > WL_POINTER_AXIS_HORIZONTAL_SCROLL) {
 		wl_resource_post_error(resource,
 				ZWLR_VIRTUAL_POINTER_V1_ERROR_INVALID_AXIS,
-				"Invalid enumeration value %d", axis);
+				"Invalid enumeration value %" PRIu32, axis);
 		return;
 	}
 	struct wlr_virtual_pointer_v1 *pointer =
@@ -189,22 +174,28 @@ static void virtual_pointer_axis_discrete(struct wl_client *client,
 	if (pointer == NULL) {
 		return;
 	}
-	struct wlr_input_device *wlr_dev = &pointer->input_device;
 	pointer->axis = axis;
 	pointer->axis_valid[pointer->axis] = true;
-	pointer->axis_event[pointer->axis].device = wlr_dev;
+	pointer->axis_event[pointer->axis].pointer = &pointer->pointer;
 	pointer->axis_event[pointer->axis].time_msec = time;
 	pointer->axis_event[pointer->axis].orientation = axis;
 	pointer->axis_event[pointer->axis].delta = wl_fixed_to_double(value);
-	pointer->axis_event[pointer->axis].delta_discrete = discrete;
+	pointer->axis_event[pointer->axis].delta_discrete = discrete *
+		WLR_POINTER_AXIS_DISCRETE_STEP;
 }
 
 static void virtual_pointer_destroy_resource(struct wl_resource *resource) {
 	struct wlr_virtual_pointer_v1 *pointer =
 		virtual_pointer_from_resource(resource);
-	if (pointer != NULL) {
-		wlr_input_device_destroy(&pointer->input_device);
+	if (pointer == NULL) {
+		return;
 	}
+
+	wlr_pointer_finish(&pointer->pointer);
+
+	wl_resource_set_user_data(pointer->resource, NULL);
+	wl_list_remove(&pointer->link);
+	free(pointer);
 }
 
 static void virtual_pointer_destroy(struct wl_client *client,
@@ -239,27 +230,19 @@ static void virtual_pointer_manager_create_virtual_pointer_with_output(
 		uint32_t id) {
 	struct wlr_virtual_pointer_manager_v1 *manager = manager_from_resource(resource);
 
-	struct wlr_virtual_pointer_v1 *virtual_pointer = calloc(1,
-		sizeof(struct wlr_virtual_pointer_v1));
+	struct wlr_virtual_pointer_v1 *virtual_pointer = calloc(1, sizeof(*virtual_pointer));
 	if (!virtual_pointer) {
 		wl_client_post_no_memory(client);
 		return;
 	}
 
-	struct wlr_pointer *pointer = calloc(1, sizeof(struct wlr_pointer));
-	if (!pointer) {
-		wlr_log(WLR_ERROR, "Cannot allocate wlr_pointer");
-		free(virtual_pointer);
-		wl_client_post_no_memory(client);
-		return;
-	}
-	wlr_pointer_init(pointer, NULL);
+	wlr_pointer_init(&virtual_pointer->pointer, &pointer_impl,
+		"wlr_virtual_pointer_v1");
 
 	struct wl_resource *pointer_resource = wl_resource_create(client,
 		&zwlr_virtual_pointer_v1_interface, wl_resource_get_version(resource),
 		id);
 	if (!pointer_resource) {
-		free(pointer);
 		free(virtual_pointer);
 		wl_client_post_no_memory(client);
 		return;
@@ -268,10 +251,6 @@ static void virtual_pointer_manager_create_virtual_pointer_with_output(
 	wl_resource_set_implementation(pointer_resource, &virtual_pointer_impl,
 		virtual_pointer, virtual_pointer_destroy_resource);
 
-	wlr_input_device_init(&virtual_pointer->input_device,
-		WLR_INPUT_DEVICE_POINTER, &input_device_impl, "virtual pointer",
-		0x0, 0x0);
-
 	struct wlr_virtual_pointer_v1_new_pointer_event event = {
 		.new_pointer = virtual_pointer,
 	};
@@ -279,7 +258,7 @@ static void virtual_pointer_manager_create_virtual_pointer_with_output(
 	if (seat) {
 		struct wlr_seat_client *seat_client =
 			wlr_seat_client_from_resource(seat);
-		event.suggested_seat = seat_client->seat;
+		event.suggested_seat = seat_client != NULL ? seat_client->seat : NULL;
 	}
 
 	if (output) {
@@ -287,12 +266,10 @@ static void virtual_pointer_manager_create_virtual_pointer_with_output(
 		event.suggested_output = wlr_output;
 	}
 
-	virtual_pointer->input_device.pointer = pointer;
 	virtual_pointer->resource = pointer_resource;
-	wl_signal_init(&virtual_pointer->events.destroy);
 
 	wl_list_insert(&manager->virtual_pointers, &virtual_pointer->link);
-	wlr_signal_emit_safe(&manager->events.new_virtual_pointer, &event);
+	wl_signal_emit_mutable(&manager->events.new_virtual_pointer, &event);
 }
 
 static void virtual_pointer_manager_create_virtual_pointer(
@@ -330,7 +307,7 @@ static void virtual_pointer_manager_bind(struct wl_client *client, void *data,
 static void handle_display_destroy(struct wl_listener *listener, void *data) {
 	struct wlr_virtual_pointer_manager_v1 *manager =
 		wl_container_of(listener, manager, display_destroy);
-	wlr_signal_emit_safe(&manager->events.destroy, manager);
+	wl_signal_emit_mutable(&manager->events.destroy, manager);
 	wl_list_remove(&manager->display_destroy.link);
 	wl_global_destroy(manager->global);
 	struct wlr_virtual_pointer_v1 *pointer, *pointer_tmp;
@@ -343,8 +320,7 @@ static void handle_display_destroy(struct wl_listener *listener, void *data) {
 
 struct wlr_virtual_pointer_manager_v1* wlr_virtual_pointer_manager_v1_create(
 		struct wl_display *display) {
-	struct wlr_virtual_pointer_manager_v1 *manager = calloc(1,
-		sizeof(struct wlr_virtual_pointer_manager_v1));
+	struct wlr_virtual_pointer_manager_v1 *manager = calloc(1, sizeof(*manager));
 	if (!manager) {
 		return NULL;
 	}
