@@ -4,7 +4,6 @@
 #include <string.h>
 #include <unistd.h>
 #include <wlr/types/wlr_data_device.h>
-#include <wlr/types/wlr_gtk_primary_selection.h>
 #include <wlr/util/log.h>
 #include <xcb/xfixes.h>
 #include "xwayland/xwm.h"
@@ -50,11 +49,12 @@ static void xwm_dnd_send_event(struct wlr_xwm *xwm, xcb_atom_t type,
 		.data = *data,
 	};
 
-	xcb_send_event(xwm->xcb_conn,
+	xwm_send_event_with_size(xwm->xcb_conn,
 		0, // propagate
 		dest->window_id,
 		XCB_EVENT_MASK_NO_EVENT,
-		(const char *)&event);
+		&event,
+		sizeof(event));
 	xcb_flush(xwm->xcb_conn);
 }
 
@@ -64,7 +64,7 @@ static void xwm_dnd_send_enter(struct wlr_xwm *xwm) {
 	struct wl_array *mime_types = &drag->source->mime_types;
 
 	xcb_client_message_data_t data = { 0 };
-	data.data32[0] = xwm->dnd_window;
+	data.data32[0] = xwm->dnd_selection.window;
 	data.data32[1] = XDND_VERSION << 24;
 
 	// If we have 3 MIME types or less, we can send them directly in the
@@ -94,7 +94,7 @@ static void xwm_dnd_send_enter(struct wlr_xwm *xwm) {
 
 		xcb_change_property(xwm->xcb_conn,
 			XCB_PROP_MODE_REPLACE,
-			xwm->dnd_window,
+			xwm->dnd_selection.window,
 			xwm->atoms[DND_TYPE_LIST],
 			XCB_ATOM_ATOM,
 			32, // format
@@ -110,7 +110,7 @@ static void xwm_dnd_send_position(struct wlr_xwm *xwm, uint32_t time, int16_t x,
 	assert(drag != NULL);
 
 	xcb_client_message_data_t data = { 0 };
-	data.data32[0] = xwm->dnd_window;
+	data.data32[0] = xwm->dnd_selection.window;
 	data.data32[2] = (x << 16) | y;
 	data.data32[3] = time;
 	data.data32[4] =
@@ -126,7 +126,7 @@ static void xwm_dnd_send_drop(struct wlr_xwm *xwm, uint32_t time) {
 	assert(dest != NULL);
 
 	xcb_client_message_data_t data = { 0 };
-	data.data32[0] = xwm->dnd_window;
+	data.data32[0] = xwm->dnd_selection.window;
 	data.data32[2] = time;
 
 	xwm_dnd_send_event(xwm, xwm->atoms[DND_DROP], &data);
@@ -139,7 +139,7 @@ static void xwm_dnd_send_leave(struct wlr_xwm *xwm) {
 	assert(dest != NULL);
 
 	xcb_client_message_data_t data = { 0 };
-	data.data32[0] = xwm->dnd_window;
+	data.data32[0] = xwm->dnd_selection.window;
 
 	xwm_dnd_send_event(xwm, xwm->atoms[DND_LEAVE], &data);
 }
@@ -151,7 +151,7 @@ static void xwm_dnd_send_leave(struct wlr_xwm *xwm) {
 	assert(dest != NULL);
 
 	xcb_client_message_data_t data = { 0 };
-	data.data32[0] = xwm->dnd_window;
+	data.data32[0] = xwm->dnd_selection.window;
 	data.data32[1] = drag->source->accepted;
 
 	if (drag->source->accepted) {
@@ -192,7 +192,7 @@ int xwm_handle_selection_client_message(struct wlr_xwm *xwm,
 		drag->source->accepted = accepted;
 		wlr_data_source_dnd_action(drag->source, action);
 
-		wlr_log(WLR_DEBUG, "DND_STATUS window=%d accepted=%d action=%d",
+		wlr_log(WLR_DEBUG, "DND_STATUS window=%" PRIu32 " accepted=%d action=%d",
 			target_window, accepted, action);
 		return 1;
 	} else if (ev->type == xwm->atoms[DND_FINISHED]) {
@@ -226,7 +226,7 @@ int xwm_handle_selection_client_message(struct wlr_xwm *xwm,
 			wlr_data_source_dnd_finish(source);
 		}
 
-		wlr_log(WLR_DEBUG, "DND_FINISH window=%d performed=%d action=%d",
+		wlr_log(WLR_DEBUG, "DND_FINISH window=%" PRIu32 " performed=%d action=%d",
 			target_window, performed, action);
 		return 1;
 	} else {
